from shy_sh.settings import settings, BaseLLMSchema
from functools import lru_cache


@lru_cache
def get_llm():
    return _get_llm(settings.llm)


def _get_llm(llm_config: BaseLLMSchema):
    llm = None
    match llm_config.provider:
        case "openai":
            from langchain_openai import ChatOpenAI

            llm = ChatOpenAI(
                model=llm_config.name,
                temperature=llm_config.temperature,
                api_key=llm_config.api_key,
                base_url=llm_config.base_url,
            )
        case "ollama":
            from langchain_ollama import ChatOllama

            llm = ChatOllama(model=llm_config.name, temperature=llm_config.temperature)

        case "groq":
            from langchain_groq import ChatGroq

            llm = ChatGroq(
                model=llm_config.name,
                temperature=llm_config.temperature,
                api_key=llm_config.api_key,
            )

        case "anthropic":
            from langchain_anthropic import ChatAnthropic

            llm = ChatAnthropic(
                model_name=llm_config.name,
                temperature=llm_config.temperature,
                anthropic_api_key=llm_config.api_key,
            )

        case "google":
            from langchain_google_genai import ChatGoogleGenerativeAI
            import os

            # Suppress logging warnings
            os.environ["GRPC_VERBOSITY"] = "ERROR"
            os.environ["GLOG_minloglevel"] = "2"

            llm = ChatGoogleGenerativeAI(
                model=llm_config.name,
                temperature=llm_config.temperature,
                api_key=llm_config.api_key,
            )

        case "aws":
            from langchain_aws import ChatBedrockConverse

            region, access_key, secret_key = llm_config.api_key.split(" ")

            llm = ChatBedrockConverse(
                model=llm_config.name,
                temperature=llm_config.temperature,
                region_name=region,
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
            )
        case _:
            raise ValueError(f"Unknown LLM provider: {llm_config.provider}")
    return llm


DEFAULT_CONTEXT_LEN = 8192
LLM_CONTEXT_WINDOWS = {
    "openai": {
        "default": DEFAULT_CONTEXT_LEN * 4,
    },
    "ollama": {
        "default": DEFAULT_CONTEXT_LEN,
    },
    "groq": {
        "default": DEFAULT_CONTEXT_LEN,
    },
    "anthropic": {
        "default": DEFAULT_CONTEXT_LEN * 4,
    },
    "google": {
        "default": DEFAULT_CONTEXT_LEN * 4,
    },
    "aws": {
        "default": DEFAULT_CONTEXT_LEN * 4,
    },
}


def get_llm_context():
    provider = LLM_CONTEXT_WINDOWS.get(settings.llm.provider, None)
    if not provider:
        return DEFAULT_CONTEXT_LEN
    return provider.get(
        settings.llm.name,
        provider.get("default", DEFAULT_CONTEXT_LEN),
    )



from typing import Annotated
from rich import print
from langgraph.prebuilt import InjectedState
from langchain.tools import tool
from shy_sh.models import State, ToolMeta
from shy_sh.utils import get_shell_history


@tool(response_format="content_and_artifact")
def shell_history(arg: str, state: Annotated[State, InjectedState]):
    """to get the history of the last commands executed by the user in the shell, useful to understand what the user has already tried, use it if the user asks to check a previous command launched before the chat started"""
    print(f"📜 [bold yellow]Let me check...[/bold yellow]\n")
    history = get_shell_history()
    return f"These are the last commands executed by the user:\n{history}", ToolMeta()



from typing import Optional
from typing_extensions import TypedDict, Annotated
from langgraph.graph.message import add_messages
from pydantic import BaseModel


class ToolRequest(BaseModel):
    tool: str
    arg: str
    thoughts: Optional[str] = None


class FinalResponse(BaseModel):
    response: str


class ToolMeta(BaseModel):
    stop_execution: bool = False
    skip_print: bool = False
    executed_scripts: list[dict] = []


def append(left: list, right: list, **kwargs) -> list:
    """Append right to left and return the result"""
    left.extend(right)
    return left


class State(TypedDict):
    timestamp: str
    lang_spec: str
    ask_before_execute: bool = True
    tools_instructions: str | None = None
    few_shot_examples: Annotated[list, add_messages] = []
    history: Annotated[list, add_messages] = []
    tool_history: Annotated[list, add_messages] = []
    executed_scripts: Annotated[list, append] = []



import os
from shy_sh.settings import settings


def capture_prompt():
    import speech_recognition as sr

    r = sr.Recognizer()
    r.energy_threshold = 4000
    r.pause_threshold = 1.5
    r.phrase_threshold = 1.5
    with sr.Microphone() as source:
        audio = r.listen(source, timeout=5)

    try:
        os.environ["GROQ_API_KEY"] = (
            os.environ.get("GROQ_API_KEY") or settings.llm.api_key
        )
        return r.recognize_groq(audio)
    except Exception:
        return ""



from shy_sh.agents.tools.shell import shell
from shy_sh.agents.tools.shell_expert import shell_expert
from shy_sh.agents.tools.python_expert import python_expert
from shy_sh.agents.tools.shell_history import shell_history

tools = [shell, shell_expert, python_expert, shell_history]
tools_by_name = {tool.name: tool for tool in tools}



from langchain_core.runnables import chain
from langchain_core.output_parsers import StrOutputParser
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from shy_sh.agents.llms import get_llm
from textwrap import dedent

msg_template = dedent(
    """
    Output only a block of python code like this:
    ```python
    [your python code]
    ```

    Write a python script that accomplishes the task. Try to avoid the usage of external libraries if not explicitly requested.
    Task: {input}
    """
)


@chain
def pyexpert_chain(_):
    llm = get_llm()
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a python expert. The current date and time is {timestamp}",
            ),
            MessagesPlaceholder("history"),
            ("human", msg_template),
        ]
    )
    return prompt | llm | StrOutputParser()



from rich import print
from shy_sh.settings import settings
from shy_sh.agents.shy_agent.graph import shy_agent_graph
from shy_sh.agents.misc import get_graph_inputs, run_few_shot_examples
from shy_sh.agents.shy_agent.audio import capture_prompt
from shy_sh.manager.history import (
    print_recent_commands,
    load_chat_history,
    print_chat_history,
    save_chat_history,
)
from shy_sh.utils import command_completer, save_history
from langchain_core.messages import HumanMessage


class ShyAgent:
    def __init__(
        self,
        interactive=False,
        ask_before_execute=True,
        audio=False,
    ):
        self.interactive = interactive
        self.ask_before_execute = ask_before_execute
        self.audio = audio
        self.history = []
        self.executed_scripts = []
        self.examples = run_few_shot_examples()
        self.chat_id = None

    def _run(self, task: str):
        self.history.append(HumanMessage(content=task))
        inputs = get_graph_inputs(
            history=self.history,
            examples=self.examples,
            ask_before_execute=self.ask_before_execute,
        )

        res = shy_agent_graph.invoke(inputs)
        self.history += res["tool_history"]
        self.executed_scripts += res["executed_scripts"]

    def _handle_command(self, command: str):
        if not command.startswith("/"):
            return False
        command = command[1:]
        match command:
            case "chats" | "c":
                print_chat_history()
                print("\n🤖: Here is the list of all chats!")
                return True
            case "history" | "h":
                print_recent_commands()
                print("\n🤖: Here are the most recently executed commands!")
                return True
            case "clear":
                self.history = []
                self.executed_scripts = []
                self.chat_id = None
                print("🤖: New chat started")
                return True
            case command if command.startswith("load "):
                try:
                    chat_id = int(command[5:])
                    self.history = load_chat_history(chat_id)
                    if self.history:
                        self.chat_id = chat_id
                        print(f"🤖: Loaded history for chat ID {chat_id}")
                    else:
                        print(f"🤖: No history found for chat ID {chat_id}")
                except ValueError:
                    print(f"🚨 [bold red]Invalid chat ID {command[5:]}[/]")
                return True
        return False

    def _load_autocomplete(self):
        try:
            import readline

            readline.set_completer(command_completer)
        except Exception:
            pass

    def _dispose_autocomplete(self):
        try:
            import readline

            readline.set_completer(lambda *args: None)
        except Exception:
            pass

    def start(self, task: str):
        if task:
            self._run(task)
            if self.history:
                save_chat_history(
                    chat_id=self.chat_id,
                    title=task if not self.chat_id else None,
                    messages=self.history,
                    executed_scripts=self.executed_scripts,
                    meta=settings.llm.model_dump(exclude={"api_key"}),
                )
        if self.interactive:
            if self.audio:
                new_task = None
                while not new_task:
                    print(f"\n🎤: ", end="")
                    new_task = capture_prompt().strip()
                    print(new_task)
            else:
                self._load_autocomplete()
                new_task = input("\n✨: ")
                self._dispose_autocomplete()
            while new_task.endswith("\\"):
                new_task = new_task[:-1] + "\n" + input("    ")
            save_history()

            if new_task == "exit" or new_task == "quit" or new_task == "q":
                print("\n🤖: 👋 Bye!\n")
                return
            if new_task.startswith("/"):
                r = self._handle_command(new_task)
                if r:
                    new_task = ""

            self.start(new_task)



from contextlib import contextmanager
from datetime import datetime
import os
from sqlalchemy import DateTime, Text, create_engine, JSON, ForeignKey
from sqlalchemy.orm import (
    declarative_base,
    sessionmaker,
    mapped_column,
    relationship,
    Mapped,
)
from enum import Enum

Base = declarative_base()


class Chat(Base):
    __tablename__ = "chat"

    id: Mapped[int] = mapped_column(primary_key=True)
    title: Mapped[str]
    messages: Mapped[list[dict]] = mapped_column(JSON)
    meta: Mapped[dict] = mapped_column(JSON)
    created_at: Mapped[datetime] = mapped_column(
        DateTime, default=lambda: datetime.utcnow()
    )

    executed_scripts: Mapped[list["ExecutedScript"]] = relationship(
        "ExecutedScript", back_populates="chat", cascade="all, delete-orphan"
    )


class ScriptType(str, Enum):
    SHELL = "shell"
    SHELL_SCRIPT = "shell_script"
    PYTHON = "python"


class ExecutedScript(Base):
    __tablename__ = "executed_script"

    id: Mapped[int] = mapped_column(primary_key=True)
    type: Mapped[ScriptType]
    script: Mapped[str] = mapped_column(Text)
    result: Mapped[str | None] = mapped_column(Text)
    created_at: Mapped[datetime] = mapped_column(
        DateTime, default=lambda: datetime.utcnow()
    )

    chat_id: Mapped[int] = mapped_column(ForeignKey("chat.id"))
    chat: Mapped[Chat] = relationship("Chat", back_populates="executed_scripts")


DATABASE_URL = os.path.expanduser("~/.config/shy/shy.db")
if not os.path.exists(os.path.dirname(DATABASE_URL)):
    os.makedirs(os.path.dirname(DATABASE_URL), exist_ok=True)
engine = create_engine(f"sqlite:///{DATABASE_URL}")
SessionLocal = sessionmaker(bind=engine)
Base.metadata.create_all(engine)


@contextmanager
def session():
    session = SessionLocal()
    try:
        yield session
        session.commit()
    except Exception as e:
        session.rollback()
        raise e
    finally:
        session.close()



import os
import yaml

from typing import Type, Any, Literal

from pydantic_settings import (
    BaseSettings,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)
from pydantic import BaseModel
from pathlib import Path
from questionary import confirm, text, select, password, Style


class BaseLLMSchema(BaseModel):
    provider: str
    name: str
    api_key: str = ""
    base_url: str | None = None
    temperature: float = 0.0


class LLMSchema(BaseLLMSchema):
    agent_pattern: Literal["function_call", "react"] = "react"


class _Settings(BaseModel):
    llm: LLMSchema = LLMSchema(provider="ollama", name="llama3.2")

    language: str = ""
    sandbox_mode: bool = False


class Settings(BaseSettings, _Settings):
    model_config = SettingsConfigDict(
        extra="ignore",
        yaml_file=[
            "~/.config/shy/config.yaml",
            "~/.config/shy/config.yml",
            "./shy.yaml",
            "./shy.yml",
        ],
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        **kwargs: Any,
    ):
        return (YamlConfigSettingsSource(settings_cls),)


settings = Settings()
PROVIDERS = ["ollama", "openai", "google", "anthropic", "groq", "aws"]


def get_or_create_settings_path():
    file_name = None
    for f in reversed(settings.model_config["yaml_file"]):  # type: ignore
        f = Path(f).expanduser()
        if os.path.exists(f):
            file_name = f
            break
    if not file_name:
        file_name = settings.model_config["yaml_file"][0]  # type: ignore
        file_name = Path(file_name).expanduser()
        os.makedirs(os.path.dirname(file_name), exist_ok=True)
    return file_name


_text_style = {
    "qmark": "",
    "style": Style.from_dict(
        {
            "selected": "fg:darkorange noreverse",
            "question": "fg:ansigreen nobold",
            "highlighted": "fg:darkorange",
            "text": "fg:darkorange",
            "answer": "fg:darkorange nobold",
            "instruction": "fg:darkorange",
        }
    ),
}

_select_style = {
    "pointer": "►",
    "instruction": " ",
    **_text_style,
}


def _try_float(x):
    try:
        float(x)
        return True
    except ValueError:
        return "Please enter a valid number"


def configure_yaml():
    provider = select(
        message="Provider:",
        choices=PROVIDERS,
        default=settings.llm.provider,
        **_select_style,
    ).unsafe_ask()
    if provider != "ollama":
        api_key = password(
            message="API Key:",
            default=settings.llm.api_key,
            **_text_style,
        ).unsafe_ask()
    else:
        api_key = settings.llm.api_key
    base_url = settings.llm.base_url
    if provider == "openai":
        base_url = (
            text(
                message="Base URL:",
                default=settings.llm.base_url or "",
                **_text_style,
            )
            .unsafe_ask()
            .strip()
        )
    base_url = base_url or None
    model = input_model(provider, api_key, settings.llm.name)
    agent_pattern = select(
        message="Agent Pattern:",
        choices=["function_call", "react"],
        default=settings.llm.agent_pattern,
        **_select_style,
    ).unsafe_ask()
    temperature = text(
        message="Temperature:",
        default=str(settings.llm.temperature),
        validate=lambda x: _try_float(x),
        **_text_style,
    ).unsafe_ask()

    llm = {
        "provider": provider,
        "name": model,
        "api_key": api_key,
        "base_url": base_url,
        "temperature": float(temperature),
        "agent_pattern": agent_pattern,
    }

    language = text("Language:", default=settings.language, **_text_style).unsafe_ask()
    sandbox_mode = confirm(
        "Sandbox Mode:",
        default=settings.sandbox_mode,
        **_text_style,
    ).unsafe_ask()

    file_name = get_or_create_settings_path()

    with open(file_name, "w") as f:
        f.write(
            yaml.dump(
                {
                    "llm": llm,
                    "language": language,
                    "sandbox_mode": sandbox_mode,
                }
            )
        )

    print(f"\nConfiguration saved to {file_name}")


def input_model(provider: str, api_key: str, default_model: str | None = None):
    try:
        match provider:
            case "ollama":
                from ollama import list

                r = list()
                model_list = [l.model.replace(":latest", "") for l in r["models"]]
                return select(
                    message="Model:",
                    choices=model_list,
                    default=default_model if default_model in model_list else None,
                    **_select_style,
                ).unsafe_ask()
            case "openai":
                from openai import OpenAI

                r = OpenAI(api_key=api_key).models.list()
                model_list = [l.id for l in r.data]
                return select(
                    message="Model:",
                    choices=model_list,
                    default=default_model if default_model in model_list else None,
                    **_select_style,
                ).unsafe_ask()
            case "google":
                from google.ai.generativelanguage import ModelServiceClient
                from google.auth.api_key import Credentials

                r = ModelServiceClient(credentials=Credentials(api_key)).list_models()
                model_list = [l.name.replace("models/", "", 1) for l in r]
                return select(
                    message="Model:",
                    choices=model_list,
                    default=default_model if default_model in model_list else None,
                    **_select_style,
                ).unsafe_ask()
            case "anthropic":
                import requests

                r = requests.get(
                    "https://api.anthropic.com/v1/models",
                    headers={"x-api-key": api_key, "anthropic-version": "2023-06-01"},
                ).json()
                model_list = [l["id"] for l in r["data"]]

                return select(
                    message="Model:",
                    choices=model_list,
                    default=default_model if default_model in model_list else None,
                    **_select_style,
                ).unsafe_ask()
            case "groq":
                from groq import Client

                r = Client(api_key=api_key).models.list()
                model_list = [l.id for l in r.data]
                return select(
                    message="Model:",
                    choices=model_list,
                    default=default_model if default_model in model_list else None,
                    **_select_style,
                ).unsafe_ask()
            case "aws":
                from boto3 import client

                region, access_key, secret_key = api_key.split(" ")
                r = client(
                    "bedrock",
                    region_name=region,
                    aws_access_key_id=access_key,
                    aws_secret_access_key=secret_key,
                ).list_foundation_models(
                    byOutputModality="TEXT",
                )
                model_list = [l["modelId"] for l in r["modelSummaries"]]
                return select(
                    message="Model:",
                    choices=model_list,
                    default=default_model if default_model in model_list else None,
                    **_select_style,
                ).unsafe_ask()
            case _:
                raise ValueError("Invalid provider")
    except Exception:
        return text(message="Model:", **_text_style).unsafe_ask()



from sqlalchemy import select, func
from sqlalchemy.orm import Session
from shy_sh.manager.sql_models import Chat, ExecutedScript, ScriptType


class ChatManager:
    def __init__(self, session: Session):
        self.session = session

    def get_chat(self, chat_id):
        return self.session.get(Chat, chat_id)

    def get_all_chats(self):
        return (
            self.session.execute(select(Chat).order_by(Chat.created_at)).scalars().all()
        )

    def get_recent_scripts(
        self,
        script_type: ScriptType | None = None,
        limit: int = 20,
    ):
        q = (
            select(
                ExecutedScript.script,
                func.count(ExecutedScript.script),
                func.max(ExecutedScript.type),
                func.max(ExecutedScript.created_at),
            )
            .group_by(ExecutedScript.script)
            .order_by(ExecutedScript.created_at.desc())
            .limit(limit)
        )
        if script_type:
            q = q.where(ExecutedScript.type == script_type)
        return self.session.execute(q).all()

    def save_chat(
        self,
        id: int | None = None,
        **kwargs,
    ):
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if id:
            chat = self.get_chat(id)
            if not chat:
                raise ValueError(f"Chat with ID {id} not found.")
            for key, value in kwargs.items():
                setattr(chat, key, value)
        else:
            chat = Chat(**kwargs)
            self.session.add(chat)

    def truncate_chats(self, keep: int = 100):
        chats = (
            self.session.execute(select(Chat).order_by(Chat.created_at)).scalars().all()
        )
        for chat in chats[:-10]:
            self.session.delete(chat)
